# -*- coding: utf-8 -*-

#   Copyright (C) 2008-2010, 2013, 2015, 2017-2018, 2020-2021,
#   2023-2025 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" Functions for working with Python frames"""

import dis
import inspect
import linecache
import os
import os.path as osp
import re
from opcode import opname
from reprlib import repr
from types import FrameType
from typing import Optional, Tuple

import xdis
from xdis import IS_PYPY, get_opcode
from xdis.version_info import PYTHON_VERSION_TRIPLE

from trepan.lib.bytecode import op_at_frame
from trepan.lib.format import (
    Arrow,
    Filename,
    Function,
    LineNumber,
    Return,
    format_python,
    format_token,
)
from trepan.lib.pp import pp
from trepan.lib.printing import printf

try:
    from trepan.processor.cmdfns import deparse_fn
except ImportError:
    def deparse_fn(code):
        raise NotImplementedError
try:
    from trepan.lib.deparse import deparse_offset

    have_deparser = True
except ImportError:

    def deparse_offset(_code, _name: str, _list_i: int, _) -> tuple:
        return None, None

    have_deparser = False

_with_local_varname = re.compile(r"_\[[0-9+]]")


def count_frames(frame: FrameType, count_start=0):
    """Return a count of the number of frames"""
    count = -count_start
    for _ in range(1000):
        if frame is None:
            break
        else:
            count += 1
            frame = frame.f_back
    else:
        return 1000
    return count


_re_pseudo_file = re.compile(r"^<.+>")


# Taken from 3.7 inspect.py. However instead of an object name, we
# start with the filename. Also, oddly getsourcefile wasn't
# stripping out __pycache__. Finally, we've adapted this
# so that it works back to 3.0.
#
def getsourcefile(filename: str) -> Optional[str]:
    """Return the filename that can be used to locate an object's source.
    Return None if no way can be identified to get the source.
    """
    all_bytecode_suffixes = (".pyc", ".pyo")
    if any(filename.endswith(s) for s in all_bytecode_suffixes):
        if osp.dirname(filename).endswith("__pycache__"):
            filename = osp.join(
                osp.dirname(osp.dirname(filename)), osp.basename(filename)
            )
        filename = osp.splitext(filename)[0] + ".py"
    elif any(filename.endswith(s) for s in [".abi3.so", ".so"]):
        return None
    if osp.exists(filename):
        return filename
    # only return a non-existent filename if the module has a PEP 302 loader
    if getattr(inspect.getmodule(object, filename), "__loader__", None) is not None:
        return filename
    # or it is in the linecache
    if filename in linecache.cache:
        return filename


def deparse_source_from_code(code):
    source_text = ""
    try:
        source_lines = deparse_fn(code).split("\n")
        source_text = ""
        for i, source_text in enumerate(source_lines):
            if len(source_text) > 0:
                break
        if len(source_lines) > 1:
            source_text += "..."
        source_text = f'"{source_text}"'
    except Exception:
        pass
    return source_text


def format_function_name(frame: FrameType, style: str) -> Tuple[Optional[str], Optional[str]]:
    """
    Pick out the function name from ``frame`` and return both the name
    and the name styled according to ``style``
    """
    if (exec_type := is_eval_or_exec_stmt(frame)):
        funcname = get_call_function_name(frame)
        if funcname is None:
            funcname = exec_type
    elif frame.f_code.co_name:
        funcname = frame.f_code.co_name
    else:
        funcname = get_call_function_name(frame)
        # funcname = "<lambda>"
        pass
    if funcname is None:
        return None, None
    return funcname, format_token(Function, funcname, style=style)


def format_function_and_parameters(frame: FrameType, debugger, style: str) -> Tuple[bool, str]:
    """ """

    funcname, s = format_function_name(frame, style)
    args, varargs, varkw, local_vars = inspect.getargvalues(frame)
    if "<module>" == funcname and (
        [],
        None,
        None,
    ) == (
        args,
        varargs,
        varkw,
    ):
        is_module = True
        if is_eval_or_exec_stmt(frame):
            fn_name = format_token(Function, "exec", style=style)
            source_text = deparse_source_from_code(frame.f_code)
            s += f" {format_token(Function, fn_name, style=style)}({source_text})"
        else:
            fn_name = get_call_function_name(frame)
            if fn_name:
                source_text = deparse_source_from_code(frame.f_code)
                if fn_name:
                    s += f" {format_token(Function, fn_name, style=style)}({source_text})"
            pass
    else:
        is_module = False
        try:
            params = inspect.formatargvalues(args, varargs, varkw, local_vars)
            formatted_params = format_python(params, style=style)
        except Exception:
            pass
        else:
            maxargstrsize = debugger.settings["maxargstrsize"]
            if len(params) >= maxargstrsize:
                params = f"{params[0:maxargstrsize]}...)"
                formatted_params = format_python(params, style=style)
                pass
            s += formatted_params
        pass

    return is_module, s


def format_return_and_location(
    frame,
    line_number: int,
    debugger,
    is_module: bool,
    include_location: bool,
    style: str,
) -> str:
    """
    Format the return value if frame is at a return, and the location
    of frame if `include_location` is True; `is_module` indicates whether
    frame is a module or not.

    A formatted string is returned; `style` gives the formatting
    style used.
    """
    filename = frame2file(debugger.core, frame)
    s = ""
    if "__return__" in frame.f_locals:
        rv = frame.f_locals["__return__"]
        s += "->"
        s += format_token(Return, repr(rv), style=style)
        pass

    if include_location:
        is_pseudo_file = _re_pseudo_file.match(filename)
        add_quotes_around_file = not is_pseudo_file
        if is_module:
            if filename == "<string>":
                s += " in exec"
            elif not is_eval_or_exec_stmt(frame) and not is_pseudo_file:
                s += " file"
        elif s == "?()":
            if is_eval_or_exec_stmt(frame):
                s = "in exec"
                # exec_str = get_exec_string(frame.f_back)
                # if exec_str != None:
                #     filename = exec_str
                #     add_quotes_around_file = False
                #     pass
                # pass
            elif not is_pseudo_file:
                s = "in file"
                pass
            pass
        elif not is_pseudo_file:
            s += " called from file"
            pass

        if add_quotes_around_file:
            filename = f"'{filename}'"
        s += " %s at line %s" % (
            format_token(Filename, filename, style=style),
            format_token(LineNumber, str(line_number), style=style),
        )

    return s


def format_stack_entry(
    dbg_obj, frame_lineno, lprefix=": ", include_location=True, style="none"
) -> str:
    """Format and return a stack entry gdb-style.
    Note: lprefix is not used. It is kept for compatibility.
    """
    frame, line_number = frame_lineno

    is_module, s = format_function_and_parameters(frame, dbg_obj, style)
    args, varargs, varkw, local_vars = inspect.getargvalues(frame)

    # Note: ddd can't handle wrapped stack entries (yet).
    # The 35 is hoaky though. FIXME.
    if len(s) >= 35:
        s += "\n    "

    s += format_return_and_location(
        frame, line_number, dbg_obj, is_module, include_location, style
    )
    return s


def frame2file(core_obj, frame, canonic=True):
    if canonic:
        return core_obj.filename(core_obj.canonic_filename(frame))
    else:
        return core_obj.filename(frame.f_code.co_filename)


def frame2filesize(frame):
    if "__cached__" in frame.f_globals:
        bc_path = frame.f_globals["__cached__"]
    else:
        bc_path = None
    path = frame.f_globals["__file__"]
    source_path = getsourcefile(path)
    if source_path is None:
        return None, None
    fs_size = os.stat(source_path).st_size
    if bc_path:
        (
            _version,
            _timestamp,
            _magic_int,
            _co,
            _is_pypy,
            bc_source_size,
            _sip_hash,
        ) = xdis.load_module(bc_path, fast_load=True, get_code=False)
        return fs_size, bc_source_size
    elif osp.exists(path):
        return fs_size, None
    else:
        return None, None


def check_path_with_frame(frame, path):
    my_size = os.stat(path).st_size
    fs_size, bc_size = frame2filesize(frame)
    if bc_size and bc_size != my_size:
        return False, "bytecode and local files mismatch"
    if fs_size and fs_size != my_size:
        return (
            False,
            (
                "frame file size, %d bytes, and local file size, %d bytes, on file %s mismatch"
                % (fs_size, my_size, path)
            ),
        )
    return True, None


def is_eval_or_exec_stmt(frame) -> Optional[str]:
    """Return "eval" or "exec" if we are inside an eval() or exec()
    statement. None is returned if not.
    """
    if not hasattr(frame, "f_back"):
        return None
    back_frame = frame.f_back
    func_name = get_call_function_name(back_frame)
    if func_name and frame.f_code.co_filename == "<string>":
        return func_name
    return None


opc = get_opcode(PYTHON_VERSION_TRIPLE, IS_PYPY)


def get_call_function_name(frame) -> Optional[str]:
    """If f_back is looking at a call function, return
    the name for it. Otherwise, return None"""

    if not frame:
        return None
    # FIXME: also handle CALL_EX, CALL_FUNCTION_KW, CALL_FUNCTION_KW_EX?
    if op_at_frame(frame) not in ("CALL_FUNCTION", "CALL"):
        return None

    co = frame.f_code
    code = co.co_code
    # labels     = dis.findlabels(code)
    linestarts = dict(dis.findlinestarts(co))
    offset = frame.f_lasti
    last_LOAD_offset = -1
    is_load_global = None
    while offset >= 0:
        opcode = code[offset]
        if opname[opcode] in ("LOAD_NAME", "LOAD_GLOBAL"):
            last_LOAD_offset = offset
            is_load_global = opname[opcode] == "LOAD_GLOBAL"
        if offset in linestarts:
            break
        offset -= 2
        pass

    if last_LOAD_offset != -1:
        offset = last_LOAD_offset + 1
        arg = code[offset]

        # FIXME: Calculate arg value with EXTENDED_ARG
        extended_arg = 0
        extended_arg_offset = last_LOAD_offset - 2
        opcode = code[extended_arg_offset]
        while extended_arg_offset >= 0 and opcode == opc.EXTENDED_ARG:
            extended_arg_offset -= 2
            opcode = code[extended_arg_offset]

        while extended_arg_offset >= 0 and opcode == opc.EXTENDED_ARG:
            extended_arg_offset += 1
            extended_arg += code[extended_arg_offset] << 8
            extended_arg_offset += 1
            opcode = code[extended_arg_offset]

        arg += extended_arg
        if is_load_global:
            arg >>= 1
        if arg < len(co.co_names):
            return co.co_names[arg]
    return None


def print_stack_entry(proc_obj, i_stack: int, style="none", opts={}):
    frame_lineno = proc_obj.stack[len(proc_obj.stack) - i_stack - 1]
    frame, lineno = frame_lineno
    intf = proc_obj.intf[-1]
    name = "??"
    if frame is proc_obj.curframe:
        intf.msg_nocr(format_token(Arrow, "->", style=style))
    else:
        intf.msg_nocr("##")
    intf.msg(
        f"{i_stack} {format_stack_entry(proc_obj.debugger, frame_lineno, style=style)}"
    )
    if opts.get("source", False):
        filename = frame2file(proc_obj.core, frame)
        line = linecache.getline(filename, lineno, frame.f_globals)
        intf.msg(line)
        pass

    if opts.get("deparse", False):
        name = frame.f_code.co_name
        if frame.f_lasti == -1:
            last_i = 0
        else:
            last_i = frame.f_lasti

        if name == "<module>":
            name = "module"

        if have_deparser:
            deparsed, node_info = deparse_offset(frame.f_code, name, last_i, None)
            if node_info:
                extract_info = deparsed.extract_node_info(node_info)
                intf.msg(extract_info.selectedLine)
                intf.msg(extract_info.markerLine)
        pass
    if opts.get("full", False):
        names = list(frame.f_locals.keys())
        for name in sorted(names):
            if _with_local_varname.match(name):
                val = frame.f_locals[name]
            else:
                val = proc_obj.getval(name, frame.f_locals)
                pass
            width = opts.get("width", 80)
            pp(val, width, intf.msg_nocr, intf.msg, prefix=f"{name} =")
            pass

        deparsed, node_info = deparse_offset(frame.f_code, name, frame.f_lasti, None)
        if name == "<module>":
            name = "module"
        if node_info:
            extract_info = deparsed.extract_node_info(node_info)
            intf.msg(extract_info.selectedLine)
            intf.msg(extract_info.markerLine)
        pass


def print_stack_trace(proc_obj, count=None, style="none", opts={}):
    "Print ``count`` entries of the stack trace"
    if count is None:
        n = len(proc_obj.stack)
    else:
        n = min(len(proc_obj.stack), count)
    try:
        for i in range(n):
            print_stack_entry(proc_obj, i, style=style, opts=opts)
    except KeyboardInterrupt:
        pass
    return


def print_dict(s, obj, title):
    if hasattr(obj, "__dict__"):
        d = obj.__dict__
        if isinstance(d, dict):
            keys = list(d.keys())
            if len(keys) == 0:
                s += f"\n  No {title}"
            else:
                s += f"\n  {title}:\n"
            keys.sort()
            for key in keys:
                s += f"    '{key}':\t{d[key]}\n"
                pass
            pass
        pass
    return s


def eval_print_obj(arg, frame, format=None, short=False):
    """Return a string representation of an object"""
    try:
        if not frame:
            # ?? Should we have set up a dummy globals
            # to have persistence?
            val = eval(arg, None, None)
        else:
            val = eval(arg, frame.f_globals, frame.f_locals)
            pass
    except Exception:
        return 'No symbol "' + arg + '" in current context.'

    return print_obj(arg, val, format, short)


def print_obj(arg, val, format=None, short=False):
    """Return a string representation of an object"""
    what = arg
    if format:
        what = format + " " + arg
        val = printf(val, format)
        pass
    s = f"{what} = {val}"
    if not short:
        s += f"\n  type = {type(val)}"
        # Try to list the members of a class.
        # Not sure if this is correct or the
        # best way to do.
        s = print_dict(s, val, "object variables")
        if hasattr(val, "__class__"):
            s = print_dict(s, val.__class__, "class variables")
            pass
        pass
    return s


# Demo stuff above
if __name__ == "__main__":

    class MockDebuggerCore:
        def canonic_filename(self, frame):
            return frame.f_code.co_filename

        def filename(self, name):
            return name

        pass

    class MockDebugger:
        def __init__(self):
            self.core = MockDebuggerCore()
            self.settings = {"maxargstrsize": 80}
            pass

        pass

    frame = inspect.currentframe()
    # print(frame2filesize(frame))
    pyc_file = osp.join(
        osp.dirname(__file__), "__pycache__", osp.basename(__file__)[:-3] + ".pyc"
    )
    # print(pyc_file, getsourcefile(pyc_file))

    from trepan.debugger import Trepan
    m = MockDebugger()

    # For testing print_stack_entry()

    dd = Trepan()
    my_frame = inspect.currentframe()
    dd.core.processor.stack = [(my_frame, 100)]
    dd.core.processor.curframe = my_frame
    # print_stack_entry(dd.core.processor, 0, "fruity")

    # print(format_stack_entry(m, (frame, 10,)))
    print(
        format_stack_entry(
            m,
            (
                frame,
                10,
            ),
            style="tango",
        )
    )
    # print("frame count: %d" % count_frames(frame))
    # print("frame count: %d" % count_frames(frame.f_back))
    # print("frame count: %d" % count_frames(frame, 1))
    # print("def statement: x=5?: %s" % repr(is_def_stmt("x=5", frame)))
    # # Not a "def" statement because frame is wrong spot
    # print(is_def_stmt("def foo():", frame))

    # def sqr(x):
    #     x * x

    def fn(x):
        frame = inspect.currentframe()
        eval_str = is_eval_or_exec_stmt(frame.f_back)
        if eval_str:
            print(f"Caller is {eval_str} stmt")
            print(format_stack_entry(dd, (frame.f_back, frame.f_back.f_code.co_firstlineno)))

        _, mess = format_function_and_parameters(frame, dd, style="tango")
        print(mess)
        print(get_call_function_name(frame))
        return

    print("=" * 30)
    fn(5)
    eval("fn(5)")
    exec("fn(5)")
    # print("=" * 30)
    # print(print_obj("fn", fn))
    # print("=" * 30)
    # print(print_obj("len", len))
    # print("=" * 30)
    # print(print_obj("MockDebugger", MockDebugger))
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2010, 2013-2015, 2018, 2023-2025
#   Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger class and top-level debugger functions.

This module contains the ``Trepan`` class and some top-level routines
for creating and invoking a debugger. Most of this module serves as:
  * a wrapper for `Debugger.core' routines,
  * a place to define debugger exceptions, and
  * debugger settings.

See also module ``cli`` which contains a command-line interface to debug
a Python script and `core' which contains the core debugging
start/stop and event-handling dispatcher and `client.py' which is a
user or client-side code for connecting to server'd debugged program.
"""

import sys
import types
from typing import Any, Callable, Union

import pyficache
import tracer
from tracer.tracefilter import TraceFilter
from xdis.load import check_object_path, load_module

from trepan.exception import DebuggerQuit, DebuggerRestart
from trepan.interfaces.user import UserInterface
from trepan.lib.core import TrepanCore

# Default settings used here
from trepan.lib.default import DEBUGGER_SETTINGS, START_OPTS
from trepan.lib.file import is_compiled_py
from trepan.lib.sighandler import SignalManager
from trepan.misc import option_set

try:
    from readline import get_line_buffer
except ImportError:

    def get_line_buffer():
        return None

    pass

debugger_obj = None



def get_code_from_pyc(filename) -> types.CodeType:
    obj_path = check_object_path(filename)
    info = load_module(obj_path)
    return info[3]


class Trepan:
    """
    Class for a top-level object.
    """

    def __init__(self, opts=dict()):
        """Create a debugger object. But depending on the value of
        key 'start' inside hash 'opts', we may or may not initially
        start debugging.

        See also ``Debugger.start`` and ``Debugger.stop``.
        """

        self.mainpyfile = None
        self.thread = None
        self.eval_string = None
        self.settings = self.DEFAULT_INIT_OPTS["settings"].copy()

        def get_option(key: str) -> Any:
            return option_set(opts, key, self.DEFAULT_INIT_OPTS)

        def completer(text: str, state):
            return self.complete(text, state)

        # set the instance variables that come directly from options.
        for opt in ("settings", "orig_sys_argv", "from_ipython"):
            setattr(self, opt, get_option(opt))
            pass

        core_opts = {}
        for opt in (
            "ignore_filter",
            "proc_opts",
            "processor",
            "step_ignore",
        ):
            core_opts[opt] = get_option(opt)
            pass

        # How are I/O for this debugger handled? This should
        # be set before calling DebuggerCore.
        interface_opts = opts.get(
            "interface_opts",
            {"debugger_name": "trepan3k", "readline": opts.get("readline")},
        )
        if "complete" not in interface_opts:
            interface_opts["complete"] = completer

        # FIXME when I pass in opts=opts things break

        inp = opts.get("input", None) if opts else None
        interface = get_option("interface") or UserInterface(
            inp=inp, opts=interface_opts
        )
        self.intf = [interface]

        out = get_option("output")
        if out:
            self.intf[-1].output = out
            pass

        self.core = TrepanCore(self, core_opts)

        # When set True, we'll also suspend our debug-hook tracing.
        # This gives us a way to prevent or allow self debugging.
        self.core.trace_hook_suspend = False

        if get_option("save_sys_argv"):
            # Save the debugged program's sys.argv? We do this so that
            # when the debugged script munges these, we have a good
            # copy to use for an exec restart
            self.program_sys_argv = list(sys.argv)
        else:
            self.program_sys_argv = None
            pass

        self.sigmgr = SignalManager(self)

        # Were we requested to activate immediately?
        if get_option("activate"):
            self.core.start(get_option("start_opts"))
            pass
        return

    def complete(self, last_token: str, state: int):
        """
        In place expansion of top-level debugger command
        for `last_token`` that we are in ``state``.
        """
        if hasattr(self.core.processor, "completer"):
            string_seen = get_line_buffer() or last_token
            results = self.core.processor.completer(string_seen, state)
            return results[state]
        return

    def run(self, cmd, start_opts=None, globals_=None, locals_=None):
        """Run debugger on string `cmd' using builtin function eval
        and if that builtin exec.  Arguments `globals_' and `locals_'
        are the dictionaries to use for local and global variables. By
        default, the value of globals is globals(), the current global
        variables. If `locals_' is not given, it becomes a copy of
        `globals_'.

        Debugger.core.start settings are passed via optional
        dictionary `start_opts'. Overall debugger settings are in
        ``Debugger.settings`` which changed after an instance is created
        . Also see `run_eval' if what you want to run is a
        run_eval'able expression have that result returned and
        `run_call' if you want to debug function run_call.
        """
        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(cmd, types.CodeType):
            self.eval_string = cmd
            cmd = cmd + "\n"
            pass
        retval = None
        self.core.start(start_opts)
        try:
            retval = eval(cmd, globals_, locals_)
        except SyntaxError:
            try:
                exec(cmd, globals_, locals_)
            except DebuggerQuit:
                pass
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return retval

    def run_exec(self, cmd, start_opts=None, globals_=None, locals_=None):
        """Run debugger on string `cmd' which will executed via the
        builtin function exec. Arguments `globals_' and `locals_' are
        the dictionaries to use for local and global variables. By
        default, the value of globals is globals(), the current global
        variables. If `locals_' is not given, it becomes a copy of
        `globals_'.

        ``Debugger.core.start`` settings are passed via optional
        dictionary `start_opts'. Overall debugger settings are in
        ``Debugger.settings`` which changed after an instance is created.
        See `run_eval' if what you want to run is an
         expression that should be ``eval``d and have that result returned.
        See ``run_call`` if you want to debug function ``run_call()``.
        """
        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(cmd, types.CodeType):
            cmd = cmd + "\n"
            pass
        self.core.start(start_opts)
        try:
            exec(cmd, globals_, locals_)
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return

    def run_call(self, func: Callable, *args, start_opts=None, **kwds):
        """Run debugger on function call: `func(*args, **kwds)'

        See also ``run_eval`` if what you want to run is an eval'able
        expression have that result returned and ``run``if you want to
        debug a statement via ``exec``.
        """
        res = None
        self.core.start(opts=start_opts)
        try:
            res = func(*args, **kwds)
        except DebuggerQuit:
            pass
        finally:
            self.core.stop()
        return res

    def run_eval(
        self,
        expr: Union[str, types.CodeType],
        start_opts=None,
        globals_=None,
        locals_=None,
    ):
        """Run debugger on string `expr' which will executed via the
        built-in Python function: eval; `globals_' and `locals_' are
        the dictionaries to use for local and global variables. If
        `globals' is not given, __main__.__dict__ (the current global
        variables) is used. If `locals_' is not given, it becomes a
        copy of `globals_'.

        See also `run_call' if what you to debug a function call and
        `run' if you want to debug general Python statements.
        """
        if not isinstance(expr, (str, types.CodeType)):
            self.intf[0].errmsg(
                "You need to pass either a string or a code type."
            )
            return

        if globals_ is None:
            globals_ = globals()
        if locals_ is None:
            locals_ = globals_
        if not isinstance(expr, types.CodeType):
            self.eval_string = expr
            expr = expr + "\n"
            pass
        retval = None
        self.core.start(start_opts)
        try:
            retval = eval(expr, globals_, locals_)
        except DebuggerQuit:
            pass
        finally:
            pyficache.remove_remap_file("<string>")
            self.core.stop()
        return retval

    def run_script(self, filename, start_opts=None, globals_=None, locals_=None):
        """Run debugger on Python script `filename'. The script may
        inspect sys.argv for command arguments. `globals_' and
        `locals_' are the dictionaries to use for local and global
        variables. If `globals' is not given, globals() (the current
        global variables) is used. If `locals_' is not given, it
        becomes a copy of `globals_'.

        True is returned if the program terminated normally and False
        if the debugger initiated a quit or the program did not normally
        terminate.

        See also `run_call' if what you to debug a function call,
        `run_eval' if you want to debug an expression, and `run' if you
        want to debug general Python statements not inside a file.
        """
        self.mainpyfile = self.core.canonic(filename)

        # Start with fresh empty copy of globals and locals and tell the script
        # that it's being run as __main__ to avoid scripts being able to access
        # the debugger namespace.
        if globals_ is None:
            import __main__  # NOQA

            globals_ = {
                "__name__": "__main__",
                "__file__": self.mainpyfile,
                "__builtins__": __builtins__,
            }  # NOQA
        if locals_ is None:
            locals_ = globals_
        retval = False
        self.core.execution_status = "Running"
        try:
            if not is_compiled_py(self.mainpyfile):
                compiled = compile(open(self.mainpyfile).read(), self.mainpyfile, "exec")
            else:
                compiled = self.mainpyfile

        except SyntaxError:
            self.intf[0].errmsg(f"Python can't compile {self.mainpyfile}")
            self.intf[0].errmsg(sys.exc_info()[1])
            retval = False
            pass
        except UnicodeDecodeError:
            self.intf[0].errmsg(
                "File %s can't be read as a text file. Is it Python source?"
                % self.mainpyfile
            )
            self.intf[0].errmsg(sys.exc_info()[1])
            retval = False
            pass
        except IOError:
            print(sys.exc_info()[1])
        except DebuggerQuit:
            retval = False
            pass
        except DebuggerRestart:
            self.core.execution_status = "Restart requested"
            raise DebuggerRestart
        else:
            if is_compiled_py(self.mainpyfile):
                compiled = get_code_from_pyc(compiled)
            self.core.start(start_opts)
            exec(compiled, globals_, locals_)
            retval = True
        finally:
            self.core.stop(options={"remove": True})
        return retval

    def restart_argv(self):
        """Return an array that would be execv-ed  to restart the program"""
        return self.orig_sys_argv or self.program_sys_argv

    # The following functions have to be defined before
    # DEFAULT_INIT_OPTS which includes references to these.

    # Note: has to come after functions listed in ignore_filter.
    ignore_items = [tracer, tracer.tracer, TrepanCore]
    trepan_debugger = sys.modules.get("trepan.debugger")
    if trepan_debugger is not None:
        ignore_items.append(trepan_debugger)
    else:
        ignore_items += [run_call, run_eval, run_script]
    DEFAULT_INIT_OPTS = {
        # What routines will we not trace into?
        "ignore_filter": TraceFilter(ignore_items),
        # sys.argv when not None contains sys.argv *before* debugger
        # command processing. So sys.argv contains debugger options as
        # well as debugged-program options. These options are used to
        # do a "hard" or execv() restart.
        # program_sys_argv is set by option save_sys_argv and contains
        # sys.argv that we see now which may have debugger options
        # stripped, or it may be that we were not called from a
        # debugger front end but from inside the running
        # program. These options are suitable for a "soft" or
        # exception-handling DebuggerRestart kind of restart.
        "orig_sys_argv": None,
        "save_sys_argv": True,
        # How is I/O for this debugger handled?
        "activate": False,
        "interface": None,
        "input": None,
        "output": None,
        "processor": None,
        # Setting contains lots of debugger settings - a whole file
        # full of them!
        "settings": DEBUGGER_SETTINGS,
        "start_opts": START_OPTS,
        "step_ignore": 0,
        "from_ipython": False,
    }

    pass


# Demo it
if __name__ == "__main__":

    def foo():
        y = 2
        for item in range(2):
            print(f"{item} {y}")
            pass
        return 3

    d = Trepan()
    print(d.settings)
    d.settings["trace"] = True
    d.settings["printset"] = tracer.ALL_EVENTS
    d.core.step_ignore = -1
    print('Issuing: run_eval("1+2")')
    print(d.run_eval("1+2"))
    print('Issuing: run_exec("x=1; y=2")')
    d.run_exec("x=1; y=2")

    print('Issuing: run("3*4")')
    print(d.run("3*4"))
    print('Issuing: run("x=3; y=4")')
    d.run("x=3; y=4")

    print("Issuing: run_call(foo)")
    d.run_call(foo)
    if len(sys.argv) > 1:
        while True:
            try:
                print("started")
                d.core.step_ignore = 0
                d.core.start()
                x = foo()
                for i in range(2):
                    print(f"{(i + 1) * 10}")
                    pass
                d.core.stop()

                def square(n):
                    return n * n

                print("calling: run_call(square,2)")
                d.run_call(square, 2)
            except DebuggerQuit:
                print("That's all Folks!...")
                break
            except DebuggerRestart:
                print("Restarting...")
                pass
            pass
        pass
    pass



# -*- coding: utf-8 -*-
# Copyright (C) 2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import inspect

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class LoadCommand(DebuggerCommand):
    """**load** *trepan3k-module*

    Load a trepan3k debugger extension

    This command is usually used by trepan3k developers working on the program

    Examples:
    --------

        load mathics3  # Loads in custom mathics3 module
    """

    short_help = "load a trepan3k command extension"

    DebuggerCommand.setup(locals(), category="support", min_args=1, max_args=1)

    def load_command(self, command_module, cmd_name: str):
        proc = self.proc
        classnames = [
            tup
            for tup in inspect.getmembers(command_module, inspect.isclass)
            if ("DebuggerCommand" != tup[0] and tup[0].endswith("Command"))
        ]
        if len(classnames) == 1:
            classname, class_obj = classnames[0]
            try:
                instance = getattr(command_module, classname)(proc)
            except Exception:
                self.errmsg(
                    f"Error loading {classname} from module name, sys.exc_info()[0]"
                )
                return

            # FIXME: should we also replace object in proc.cmd_instances?
            proc.commands[cmd_name] = instance
            for alias in class_obj.aliases:
                if alias not in proc.aliases:
                    proc.aliases[alias] = cmd_name
            if hasattr(command_module, "setup") and inspect.isfunction(
                command_module.setup
            ):
                command_module.setup(self.debugger, instance)

            self.msg(f'loaded command: "{cmd_name}"')

    def run(self, args):
        module_name = args[1]
        cmd_name_array = module_name.split(".")
        try:
            command_module = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            self.errmsg(str(e))
            return

        if len(cmd_name_array) > 1:
            cmd_name = cmd_name_array[-1]
            self.load_command(command_module, cmd_name)
        else:
            modules = inspect.getmembers(command_module, inspect.ismodule)
            for cmd_name, module in modules:
                if cmd_name.startswith("__"):
                    continue
                self.load_command(module, cmd_name)
        return

    pass


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    dbgr, cmd = Mmock.dbg_setup()
    command = LoadCommand(cmd)
    for cmdline in [
        "load trepan3k_mathics3",
        "load trepan3k_mathics3.mup",
        "load trepan3k_mathics3.mbacktrace",
    ]:
        args = cmdline.split()
        cmd_argstr = cmdline[len(args[0]) :].lstrip()
        cmd.cmd_argstr = cmd_argstr
        command.run(args)
        pass
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import linecache
import os.path as osp
import re
import sys
from inspect import ismodule, currentframe
from tempfile import NamedTemporaryFile
from types import CodeType

import pyficache

from trepan.lib.stack import check_path_with_frame, frame2file, is_eval_or_exec_stmt
from trepan.processor import cmdfns
from trepan.processor.cmdfns import deparse_fn

from trepan.lib.format import (  # Opcode,
    Filename,
    Hex,
    LineNumber,
    Symbol,
    format_token,
)

try:
    from trepan.lib.deparse import deparse_and_cache

    have_deparse_and_cache = True
except ImportError:
    have_deparse_and_cache = False
    pass

warned_file_mismatches = set()


def format_code(code_object: CodeType, style) -> str:
    """
    Format according to "style" a Python code object. The
    formatted string is returned.
    """
    formatted_line = format_token(LineNumber, str(code_object.co_firstlineno), style=style)
    formatted_id = format_token(Hex, hex(id(code_object)), style=style)
    formatted_name = format_token(Symbol, code_object.co_name, style=style)
    formatted_filename = format_token(Filename, code_object.co_filename, style=style)
    return (
        f"<code object {formatted_name} at {formatted_id} "
        f"file {formatted_filename}, line {formatted_line}>"
    )


def format_frame(frame_object, style) -> str:
    """
    Format according to "style" a Python frame object. The
    formatted string is returned.
    """
    formatted_line = format_token(LineNumber, str(frame_object.f_lineno), style=style)
    formatted_id = format_token(Hex, hex(id(frame_object)), style=style)
    formatted_filename = format_token(Filename, frame_object.f_code.co_filename, style=style)
    return (
        f"<frame at {formatted_id} "
        f"file {formatted_filename}, line {formatted_line}>"
    )


def print_source_line(msg, lineno, line, event_str=None):
    """Print out a source line of text , e.g. the second
    line in:
        (/tmp.py:2):  <module>
        L -- 2 import sys,os
        (trepan3k)

    We define this method
    specifically so it can be customized for such applications
    like ipython."""

    # We don't use the filename normally. ipython and other applications
    # however might.
    return msg(f"{event_str} {lineno} {line}")


def print_source_location_info(
    print_fn, filename, lineno, fn_name=None, f_lasti=None, remapped_file=None
):
    """Print out a source location , e.g. the first line in
    line in:
        (/tmp.py:2 @21):  <module>
        L -- 2 import sys,os
        (trepan3k)
    """
    if remapped_file and filename != remapped_file:
        mess = f"({remapped_file}:{lineno} remapped {filename}"
    else:
        mess = f"({filename}:{lineno}"
    if f_lasti and f_lasti != -1:
        mess += " @%d" % f_lasti
        pass
    mess += "):"
    if fn_name and fn_name != "?":
        mess += f" {fn_name}"
        pass
    print_fn(mess)
    return

def print_location(proc_obj):
    """Show where we are. GUI's and front-end interfaces often
    use this to update displays. So it is helpful to make sure
    we give at least some place that's located in a file.
    """

    def prefix_for_filename(deparsed_text: str) -> str:
        """
        Return a reasonable filename-friendly string from some
        deparsed text. We do this so that the user gets some idea of
        what the string (source code text) is contained in the file.

        """
        lines = deparsed_text.split("\n")
        # FIXME Rather than blindly take the first line,
        # check if it is blank and if so use other lines.
        first_text_line = lines[0]
        return proc_obj._saferepr(first_text_line)[1:-1][:10]

    i_stack = proc_obj.curindex
    if i_stack is None or proc_obj.stack is None:
        return False
    core_obj = proc_obj.core
    dbgr_obj = proc_obj.debugger
    intf_obj = dbgr_obj.intf[-1]

    # Evaluation routines like "exec" don't show useful location
    # info. In these cases, we will use the position before that in
    # the stack.  Hence the looping below which in practices loops
    # once and sometimes twice.
    remapped_file = None
    source_text = None
    eval_kind = None
    i_stack = min(i_stack, len(proc_obj.stack) - 1)
    while i_stack >= 0:
        frame, lineno = proc_obj.stack[i_stack]

        # Before starting a program a location for a module with
        # line number 0 may be reported. Treat that as though
        # we were on the first line.
        if frame.f_code.co_name == "<module>" and lineno == 0:
            lineno = 1

        i_stack -= 1

        #         # Next check to see that local variable breadcrumb exists and
        #         # has the magic dynamic value.
        #         # If so, it's us and we don't normally show this.a
        #         if 'breadcrumb' in frame.f_locals:
        #             if self.run == frame.f_locals['breadcrumb']:
        #                 break

        filename = frame2file(core_obj, frame, canonic=False)
        if "<string>" == filename and dbgr_obj.eval_string:
            eval_kind = "string"
            remapped_file = filename
            filename = pyficache.unmap_file(filename)
            if "<string>" == filename:
                remapped_file = cmdfns.source_tempfile_remap(
                    "eval_string",
                    dbgr_obj.eval_string,
                    tempdir=proc_obj.settings("tempdir"),
                )
                pyficache.remap_file(filename, remapped_file)
                filename, lineno = pyficache.unmap_file_line(filename, lineno)
                pass
            pass
        elif "<string>" == filename:
            # FIXME: should change filename to disambiguated <string> everywhere.
            eval_kind = is_eval_or_exec_stmt(frame) or "code-"
            deparsed = deparse_fn(frame.f_code)
            if deparsed:
                # Create a nice prefix for the temporary file to write.
                # Use the exec type and first line of the deparsed text.
                leading_code_str = prefix_for_filename(deparsed.text)
                prefix = f"{eval_kind}-{leading_code_str}-"

                remapped_file = cmdfns.source_tempfile_remap(
                    prefix,
                    deparsed.text,
                    tempdir=proc_obj.settings("tempdir"),
                )
                # FIXME: pyficache remaps seems backwards
                pyficache.remap_file(remapped_file, filename)
                filename = remapped_file
            else:
                eval_kind = is_eval_or_exec_stmt(frame)
                deparsed = deparse_fn(frame.f_code)
                if deparsed is not None:
                    source_text = deparsed.text
                # else:
                #   print("Can't deparse", frame.f_code)
                if source_text is None and eval_kind:
                    source_text = f"{eval_kind}(...)"
                    pass
                pass
            pass
        else:
            m = re.search("^<frozen (.*)>", filename)
            if m and m.group(1) in pyficache.file2file_remap:
                remapped_file = pyficache.file2file_remap[m.group(1)]
                pass
            elif filename in pyficache.file2file_remap:
                remapped_file = pyficache.file2file_remap[filename]
                filename = remapped_file
                pass
            elif pyficache.main.remap_re_hash:
                remapped_file = pyficache.remap_file_pat(
                    filename, pyficache.main.remap_re_hash
                )
            elif m and m.group(1) in sys.modules:
                remapped_file = m.group(1)
                pyficache.remap_file(filename, remapped_file)
            pass

        opts = {
            "reload_on_change": proc_obj.settings("reload"),
            "output": proc_obj.settings("highlight"),
        }

        if proc_obj.debugger.settings.get("highlight", "plain") == "plain":
            opts["style"] = "plain"
        elif "style" in proc_obj.debugger.settings:
            opts["style"] = proc_obj.settings("style")

        pyficache.update_cache(filename)
        line = pyficache.getline(filename, lineno, opts)
        if not line:
            if (
                not source_text
                and filename.startswith("<string: ")
                and proc_obj.curframe.f_code
                and have_deparse_and_cache
            ):
                # Deparse the code object into a temp file and remap the line from code
                # into the corresponding line of the tempfile
                co = proc_obj.curframe.f_code
                tempdir = proc_obj.settings("tempdir")
                temp_filename, _ = deparse_and_cache(
                    co, proc_obj.errmsg, tempdir=tempdir
                )
                lineno = 1
                # _, lineno = pyficache.unmap_file_line(temp_filename, lineno, True)
                if temp_filename:
                    filename = temp_filename
                pass

            else:
                # FIXME:
                if source_text:
                    lines = source_text.split("\n")
                    temp_name = "string-" + prefix_for_filename(source_text)
                else:
                    # try with good ol linecache and consider fixing pyficache
                    lines = linecache.getlines(filename)
                    temp_name = filename
                if lines and not filename.endswith(".pyasm"):
                    # FIXME: DRY code with version in cmdproc.py print_location
                    prefix = osp.basename(temp_name).split(".")[0]
                    fd = NamedTemporaryFile(
                        suffix=".py",
                        prefix=prefix,
                        delete=False,
                        dir=proc_obj.settings("tempdir"),
                    )
                    with fd:
                        fd.write("".join(lines).encode("utf-8"))
                        remapped_file = fd.name
                        pyficache.remap_file(remapped_file, filename)
                    fd.close()
                    intf_obj.msg(f"remapped file {filename} to {remapped_file}")

                    pass

            if isinstance(proc_obj.curframe, int):
                breakpoint()
            line = linecache.getline(filename, lineno, proc_obj.curframe.f_globals)
            if not line:
                m = re.search("^<frozen (.*)>", filename)
                if m and m.group(1):
                    remapped_file = m.group(1)
                    try_module = sys.modules.get(remapped_file)
                    if (
                        try_module
                        and ismodule(try_module)
                        and hasattr(try_module, "__file__")
                    ):
                        remapped_file = sys.modules[remapped_file].__file__
                        pyficache.remap_file(filename, remapped_file)
                        line = linecache.getline(
                            remapped_file, lineno, proc_obj.curframe.f_globals
                        )
                    else:
                        remapped_file = m.group(1)
                        code = proc_obj.curframe.f_code
                        filename, line = cmdfns.deparse_getline(
                            code, remapped_file, lineno, opts
                        )
                    pass
            pass

        if eval_kind is None:
            fn_name = frame.f_code.co_name
            try:
                match, reason = check_path_with_frame(frame, filename)
                if not match:
                    if filename not in warned_file_mismatches:
                        proc_obj.errmsg(reason)
                        warned_file_mismatches.add(filename)
            except Exception:
                pass
            pass
        else:
            fn_name = eval_kind

        last_i = frame.f_lasti
        print_source_location_info(
            intf_obj.msg,
            filename,
            lineno,
            fn_name,
            remapped_file=remapped_file,
            f_lasti=last_i,
        )
        if line and len(line.strip()) != 0:
            if proc_obj.event:
                print_source_line(
                    intf_obj.msg, lineno, line, proc_obj.event2short[proc_obj.event]
                )
            pass
        if "<string>" != filename:
            break
        pass

    if proc_obj.event in ["return", "exception"]:
        val = proc_obj.event_arg
        intf_obj.msg(f"R=> {proc_obj._saferepr(val)}")
        pass
    elif (
        proc_obj.event == "call"
        and proc_obj.curframe.f_locals.get("__name__", "") != "__main__"
    ):
        try:
            proc_obj.commands["info"].run(["info", "locals"])
        except Exception:
            pass
    return True

# Demo it
if __name__ == "__main__":
    def five():
        from trepan.processor.cmdproc import CommandProcessor
        from trepan.processor.command.mock import MockDebugger
        d = MockDebugger()
        cmdproc = CommandProcessor(d.core)
        cmdproc.frame = currentframe()
        cmdproc.event = "line"
        cmdproc.setup()
        print_location(cmdproc)
        cmdproc.curindex = 1
        cmdproc.curframe = cmdproc.stack[cmdproc.curindex][0]
        print_location(cmdproc)

    exec("five()")



#!/usr/bin/env python3
"""Greatest Common Divisor

Some characterstics of this program used for testing check_args() does
not have a 'return' statement.

check_args() raises an uncaught exception when given the wrong number
of parameters.

"""
import sys


def check_args():
    if len(sys.argv) != 3:
        # Rather than use sys.exit let's just raise an error
        raise Exception(f"Need to give two numbers; got: {sys.argv}")
    for i in range(2):
        try:
            sys.argv[i + 1] = int(sys.argv[i + 1])
        except ValueError:
            print(f"** Expecting an integer, got: {repr(sys.argv[i])}")
            sys.exit(2)
        pass


def gcd(a, b):
    """GCD. We assume positive numbers"""

    # Make: a <= b
    if a > b:
        (a, b) = (b, a)
        pass

    if a <= 0:
        return None
    if a == 1 or b - a == 0:
        return a
    return gcd(b - a, a)


if __name__ == "__main__":
    check_args()

    (a, b) = sys.argv[1:3]
    print(f"The GCD of {a} and {b} is {gcd(a, b)}")
    pass
